from typing import List

from docarray import Document, DocumentArray
from fastapi import APIRouter
from jina import Client

from deployment.bff.app.v1.models.text import (
    NowTextIndexRequestModel,
    NowTextResponseModel,
    NowTextSearchRequestModel,
)
from deployment.bff.app.v1.routers.helper import process_query

router = APIRouter()


# Index
@router.post(
    "/index",
    summary='Add more data to the indexer',
)
def index(data: NowTextIndexRequestModel):
    """
    Append the list of texts to the indexer.
    """
    index_docs = DocumentArray()
    for text in data.texts:
        index_docs.append(Document(text=text))

    if 'wolf.jina.ai' in data.host:
        c = Client(host=data.host)
    else:
        c = Client(host=data.host, port=data.port)
    c.post('/index', index_docs)


# Search
@router.post(
    "/search",
    response_model=List[NowTextResponseModel],
    summary='Search text data via text or image as query',
)
def search(data: NowTextSearchRequestModel):
    """
    Retrieve matching texts for a given text as query. Query should be `base64` encoded
    using human-readable characters - `utf-8`.
    """
    query_doc = process_query(data.text, data.image)
    if 'wolf.jina.ai' in data.host:
        c = Client(host=data.host)
    else:
        c = Client(host=data.host, port=data.port)
    docs = c.post('/search', query_doc, parameters={"limit": data.limit})
    return docs[0].matches.to_dict()



# -*- coding: utf-8 -*-
"""
`checkbox` type question
"""
from prompt_toolkit.application import Application
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.filters import IsDone
from prompt_toolkit.layout.controls import FormattedTextControl
from prompt_toolkit.layout.containers import ConditionalContainer, \
    ScrollOffsets, HSplit, Window, WindowAlign
from prompt_toolkit.layout.dimension import LayoutDimension as D

from prompt_toolkit.layout import Layout


from . import PromptParameterException
from ..separator import Separator
from .common import setup_simple_validator, default_style, if_mousedown


# custom control based on FormattedTextControl


class InquirerControl(FormattedTextControl):
    def __init__(self, choices, pointer_index, **kwargs):
        self.pointer_index = pointer_index
        self.pointer_sign = kwargs.pop("pointer_sign", "\u276f")
        self.selected_sign = kwargs.pop("selected_sign", "\u25cf")
        self.unselected_sign = kwargs.pop("unselected_sign", "\u25cb")
        self.selected_options = []  # list of names
        self.answered = False
        self.answered_correctly = True
        self.error_message = None
        self._init_choices(choices)
        super().__init__(self._get_choice_tokens, **kwargs)

    def _init_choices(self, choices):
        # helper to convert from question format to internal format
        self.choices = []  # list (name, value)
        searching_first_choice = True if self.pointer_index == 0 else False
        for i, c in enumerate(choices):
            if isinstance(c, Separator):
                self.choices.append(c)
            else:
                name = c['name']
                value = c.get('value', name)
                disabled = c.get('disabled')
                description = c.get('description', None)
                if c.get('checked') and not disabled:
                    self.selected_options.append(value)
                self.choices.append((name, value, disabled, description))
                if searching_first_choice and not disabled:  # find the first (available) choice
                    self.pointer_index = i
                    searching_first_choice = False

    @property
    def choice_count(self):
        return len(self.choices)

    def _get_choice_tokens(self):
        tokens = []

        def append(index, line):
            if isinstance(line, Separator):
                tokens.append(('class:Separator', '  %s\n' % line))
            else:
                line_name = line[0]
                line_value = line[1]
                selected = (line_value in self.selected_options)  # use value to check if option has been selected
                pointed_at = (index == self.pointer_index)

                @if_mousedown
                def select_item(mouse_event):
                    # bind option with this index to mouse event
                    if line_value in self.selected_options:
                        self.selected_options.remove(line_value)
                    else:
                        self.selected_options.append(line_value)

                if pointed_at:
                    tokens.append(('class:pointer', ' {}'.format(self.pointer_sign), select_item))  # ' >'
                else:
                    tokens.append(('', '  ', select_item))
                # 'o ' - FISHEYE
                if choice[2]:  # disabled
                    tokens.append(('', '- %s (%s)' % (choice[0], choice[2])))
                else:
                    if selected:
                        tokens.append(('class:selected', '{} '.format(self.selected_sign), select_item))
                    else:
                        tokens.append(('', '{} '.format(self.unselected_sign), select_item))

                    if pointed_at:
                        tokens.append(('[SetCursorPosition]', ''))

                    if choice[3]:  # description
                        tokens.append(('', "%s - %s" % (line_name, choice[3])))
                    else:
                        tokens.append(('', line_name, select_item))
                tokens.append(('', '\n'))

        # prepare the select choices
        for i, choice in enumerate(self.choices):
            append(i, choice)
        tokens.pop()  # Remove last newline.
        return tokens

    def get_selected_values(self):
        # get values not labels
        return [c[1] for c in self.choices if not isinstance(c, Separator) and
                c[1] in self.selected_options]

    @property
    def line_count(self):
        return len(self.choices)


def question(message, **kwargs):
    # TODO add bottom-bar (Move up and down to reveal more choices)
    # TODO extract common parts for list, checkbox, rawlist, expand
    # TODO validate
    if not 'choices' in kwargs:
        raise PromptParameterException('choices')
    # this does not implement default, use checked...
    if 'default' in kwargs:
        raise ValueError('Checkbox does not implement \'default\' '
                         'use \'checked\':True\' in choice!')

    choices = kwargs.pop('choices', None)
    validator = setup_simple_validator(kwargs)

    # TODO style defaults on detail level
    style = kwargs.pop('style', default_style)

    pointer_index = kwargs.pop('pointer_index', 0)
    additional_parameters = dict()
    additional_parameters.update({"pointer_sign": kwargs.pop('pointer_sign', '\u276f')})
    additional_parameters.update({"selected_sign": kwargs.pop('selected_sign', '\u25cf')})
    additional_parameters.update({"unselected_sign": kwargs.pop('unselected_sign', '\u25cb')})

    ic = InquirerControl(choices, pointer_index, **additional_parameters)
    qmark = kwargs.pop('qmark', '?')

    def get_prompt_tokens():
        tokens = []

        tokens.append(('class:questionmark', qmark))
        tokens.append(('class:question', ' %s ' % message))
        if ic.answered:
            nbr_selected = len(ic.selected_options)
            if nbr_selected == 0:
                tokens.append(('class:answer', ' done'))
            elif nbr_selected == 1:
                tokens.append(('class:Answer', ' [%s]' % ic.selected_options[0]))
            else:
                tokens.append(('class:answer',
                               ' done (%d selections)' % nbr_selected))
        else:
            tokens.append(('class:instruction',
                           ' (<up>, <down> to move, <space> to select, <a> '
                           'to toggle, <i> to invert)'))
            if not ic.answered_correctly:
                tokens.append((Token.Error, ' Error: %s' % ic.error_message))
        return tokens

    # assemble layout
    layout = HSplit([
        Window(height=D.exact(1),
               content=FormattedTextControl(get_prompt_tokens),
               align=WindowAlign.CENTER,
        ),
        ConditionalContainer(
            Window(
                ic,
                width=D.exact(43),
                height=D(min=3),
                scroll_offsets=ScrollOffsets(top=1, bottom=1)
            ),
            filter=~IsDone()
        )
    ])

    # key bindings
    kb = KeyBindings()

    @kb.add('c-q', eager=True)
    @kb.add('c-c', eager=True)
    def _(event):
        raise KeyboardInterrupt()
        # event.app.exit(result=None)

    @kb.add(' ', eager=True)
    def toggle(event):
        pointed_choice = ic.choices[ic.pointer_index][1]  # value
        if pointed_choice in ic.selected_options:
            ic.selected_options.remove(pointed_choice)
        else:
            ic.selected_options.append(pointed_choice)

    @kb.add('i', eager=True)
    def invert(event):
        inverted_selection = [c[1] for c in ic.choices if
                              not isinstance(c, Separator) and
                              c[1] not in ic.selected_options and
                              not c[2]]
        ic.selected_options = inverted_selection

    @kb.add('a', eager=True)
    def all(event):
        all_selected = True  # all choices have been selected
        for c in ic.choices:
            if not isinstance(c, Separator) and c[1] not in ic.selected_options and not c[2]:
                # add missing ones
                ic.selected_options.append(c[1])
                all_selected = False
        if all_selected:
            ic.selected_options = []

    @kb.add('down', eager=True)
    def move_cursor_down(event):
        def _next():
            ic.pointer_index = ((ic.pointer_index + 1) % ic.line_count)
        _next()
        while isinstance(ic.choices[ic.pointer_index], Separator) or \
                ic.choices[ic.pointer_index][2]:
            _next()

    @kb.add('up', eager=True)
    def move_cursor_up(event):
        def _prev():
            ic.pointer_index = ((ic.pointer_index - 1) % ic.line_count)
        _prev()
        while isinstance(ic.choices[ic.pointer_index], Separator) or \
                ic.choices[ic.pointer_index][2]:
            _prev()

    @kb.add('enter', eager=True)
    def set_answer(event):
        ic.answered = True
        # TODO use validator
        event.app.exit(result=ic.get_selected_values())

    return Application(
        layout=Layout(layout),
        key_bindings=kb,
        mouse_support=True,
        style=style
    )



"""
This module implements a command-line dialog with the user.
Its goal is to configure a UserInput object with users specifications.
Optionally, values can be passed from the command-line when jina-now is launched. In that case,
the dialog won't ask for the value.
"""
from __future__ import annotations, print_function, unicode_literals

import os
import pathlib
from dataclasses import dataclass
from os.path import expanduser as user
from typing import Dict, List, Optional, Union

import cowsay
from kubernetes import client, config
from pyfiglet import Figlet

from now.constants import (
    AVAILABLE_DATASET,
    IMAGE_MODEL_QUALITY_MAP,
    DatasetTypes,
    Modalities,
    Qualities,
)
from now.deployment.deployment import cmd
from now.log.log import yaspin_extended
from now.thirdparty.PyInquirer import Separator
from now.thirdparty.PyInquirer.prompt import prompt
from now.utils import ffmpeg_is_installed, sigmap

cur_dir = pathlib.Path(__file__).parent.resolve()
NEW_CLUSTER = {'name': '🐣 create new', 'value': 'new'}
AVAILABLE_SOON = 'will be available in upcoming versions'


@dataclass
class UserInput:
    output_modality: Optional[Modalities] = None

    # data related
    data: Optional[str] = None
    is_custom_dataset: Optional[bool] = None

    custom_dataset_type: Optional[DatasetTypes] = None
    dataset_secret: Optional[str] = None
    dataset_url: Optional[str] = None
    dataset_path: Optional[str] = None

    # model related
    quality: Optional[Qualities] = None
    model_variant: Optional[str] = None

    # cluster related
    cluster: Optional[str] = None
    create_new_cluster: Optional[bool] = False
    deployment_type: Optional[str] = None


def configure_user_input(**kwargs) -> UserInput:
    print_headline()

    user_input = UserInput()
    _configure_output_modality(user_input, **kwargs)
    _configure_dataset(user_input, **kwargs)
    _configure_quality(user_input, **kwargs)
    _configure_cluster(user_input, **kwargs)

    return user_input


def print_headline():
    f = Figlet(font='slant')
    print('Welcome to:')
    print(f.renderText('Jina NOW'))
    print('Get your search case up and running - end to end.\n')
    print(
        'You can choose between image and text search. \nJina now trains a model, pushes it to the jina hub'
        ', deploys a flow and a playground app in the cloud or locally. \nCheckout one of the demo cases or bring '
        'your own data.\n'
    )
    print(
        'If you want learn more about our framework please visit: https://docs.jina.ai/'
    )
    print(
        '💡 Make sure you give enough memory to your Docker daemon. '
        '5GB - 8GB should be okay.'
    )
    print()


def _configure_output_modality(user_input: UserInput, **kwargs) -> None:
    """Asks user questions to set output_modality in user_input"""
    user_input.output_modality = _prompt_value(
        name='output_modality',
        choices=[
            {'name': '🏞 Image Search', 'value': Modalities.IMAGE},
            {'name': '📝 Text Search (experimental)', 'value': Modalities.TEXT},
            {
                'name': '🥁 Music Search',
                'value': Modalities.MUSIC,
                'disabled': AVAILABLE_SOON,
            },
        ],
        prompt_message='Which modalities you want to work with?',
        prompt_type='list',
        **kwargs,
    )


def _configure_dataset(user_input: UserInput, **kwargs) -> None:
    """Asks user to set dataset attribute of user_input"""
    if user_input.output_modality == Modalities.IMAGE:
        _configure_dataset_image(user_input, **kwargs)
    elif user_input.output_modality == Modalities.TEXT:
        _configure_dataset_text(user_input, **kwargs)
    elif user_input.output_modality == Modalities.MUSIC:
        if not ffmpeg_is_installed():
            _handle_ffmpeg_install_required()
        _configure_dataset_music(user_input, **kwargs)

    if user_input.data in AVAILABLE_DATASET[user_input.output_modality]:
        user_input.is_custom_dataset = False
    else:
        user_input.is_custom_dataset = True
        if user_input.data == 'custom':
            _configure_custom_dataset(user_input, **kwargs)
        else:
            _parse_custom_data_from_cli(user_input)


def _configure_dataset_image(user_input: UserInput, **kwargs) -> None:
    user_input.data = _prompt_value(
        name='data',
        prompt_message='What dataset do you want to use?',
        choices=[
            {'name': '🖼  artworks (≈8K docs)', 'value': 'best-artworks'},
            {
                'name': '💰 nft - bored apes (10K docs)',
                'value': 'nft-monkey',
            },
            {'name': '👬 totally looks like (≈12K docs)', 'value': 'tll'},
            {'name': '🦆 birds (≈12K docs)', 'value': 'bird-species'},
            {'name': '🚗 cars (≈16K docs)', 'value': 'stanford-cars'},
            {
                'name': '🏞 geolocation (≈50K docs)',
                'value': 'geolocation-geoguessr',
            },
            {'name': '👕 fashion (≈53K docs)', 'value': 'deepfashion'},
            {
                'name': '☢️ chest x-ray (≈100K docs)',
                'value': 'nih-chest-xrays',
            },
            Separator(),
            {
                'name': '✨ custom',
                'value': 'custom',
            },
        ],
        **kwargs,
    )


def _configure_dataset_text(user_input: UserInput, **kwargs):
    user_input.data = _prompt_value(
        name='data',
        prompt_message='What dataset do you want to use?',
        choices=[
            {'name': '🎤 rock lyrics (200K docs)', 'value': 'rock-lyrics'},
            {'name': '🎤 pop lyrics (200K docs)', 'value': 'pop-lyrics'},
            {'name': '🎤 rap lyrics (200K docs)', 'value': 'rap-lyrics'},
            {'name': '🎤 indie lyrics (200K docs)', 'value': 'indie-lyrics'},
            {'name': '🎤 metal lyrics (200K docs)', 'value': 'metal-lyrics'},
            Separator(),
            {
                'name': '✨ custom .txt files',
                'value': 'custom',
            },
        ],
        **kwargs,
    )


def _configure_dataset_music(user_input: UserInput, **kwargs):
    user_input.data = _prompt_value(
        name='data',
        prompt_message='What dataset do you want to use?',
        choices=[
            {'name': '🎸 music small (≈2K docs)', 'value': 'music-genres-small'},
            {'name': '🎸 music large (≈10K docs)', 'value': 'music-genres-large'},
            Separator(),
            {
                'name': '✨ custom',
                'value': 'custom',
            },
        ],
        **kwargs,
    )


def _configure_custom_dataset(user_input: UserInput, **kwargs) -> None:
    """Asks user questions to setup custom dataset in user_input."""
    user_input.custom_dataset_type = _prompt_value(
        name='custom_dataset_type',
        prompt_message='How do you want to provide input? (format: https://docarray.jina.ai/)',
        choices=[
            {
                'name': 'docarray.pull id (recommended)',
                'value': 'docarray',
            },
            {
                'name': 'docarray URL',
                'value': 'url',
            },
            {
                'name': 'local path',
                'value': 'path',
            },
        ],
        **kwargs,
    )

    if user_input.custom_dataset_type == 'docarray':
        user_input.dataset_secret = _prompt_value(
            name='dataset_secret',
            prompt_message='Please enter your docarray secret.',
            prompt_type='password',
        )
    elif user_input.custom_dataset_type == 'url':
        user_input.dataset_url = _prompt_value(
            name='dataset_url',
            prompt_message='Please paste in your url for the docarray.',
            prompt_type='input',
        )
    elif user_input.custom_dataset_type == 'path':
        user_input.dataset_path = _prompt_value(
            name='dataset_path',
            prompt_message='Please enter the path to the local folder.',
            prompt_type='input',
        )


def _configure_cluster(user_input: UserInput, skip=False, **kwargs):
    """Asks user question to determine cluster for user_input object"""

    def ask_deployment():
        user_input.deployment_type = _prompt_value(
            name='deployment_type',
            choices=[
                {
                    'name': '⛅️ Jina Cloud',
                    'value': 'remote',
                },
                {
                    'name': '📍 Local',
                    'value': 'local',
                },
            ],
            prompt_message='Where do you want to deploy your search engine?',
            prompt_type='list',
            **kwargs,
        )

    if not skip:
        ask_deployment()

    if user_input.deployment_type == 'remote':
        _maybe_login_wolf()
        os.environ['JCLOUD_NO_SURVEY'] = '1'
    else:
        # get all local cluster contexts
        choices = _construct_local_cluster_choices(
            active_context=kwargs.get('active_context'), contexts=kwargs.get('contexts')
        )
        # prompt the local cluster context choices to the user
        user_input.cluster = _prompt_value(
            name='cluster',
            choices=choices,
            prompt_message='Which cluster you want to use to deploy your search engine?',
            prompt_type='list',
            **kwargs,
        )
        if user_input.cluster != NEW_CLUSTER['value']:
            if not _cluster_running(user_input.cluster):
                print(
                    f'Cluster {user_input.cluster} is not running. Please select a different one.'
                )
                _configure_cluster(user_input, skip=True, **kwargs)
        else:
            user_input.create_new_cluster = True


def _configure_quality(user_input: UserInput, **kwargs) -> None:
    """Asks users questions to set quality attribute of user_input"""
    if user_input.output_modality == Modalities.MUSIC:
        return
    user_input.quality = _prompt_value(
        name='quality',
        choices=[
            {'name': '🦊 medium (≈3GB mem, 15q/s)', 'value': Qualities.MEDIUM},
            {'name': '🐻 good (≈3GB mem, 2.5q/s)', 'value': Qualities.GOOD},
            {
                'name': '🦄 excellent (≈4GB mem, 0.5q/s)',
                'value': Qualities.EXCELLENT,
            },
        ],
        prompt_message='What quality do you expect?',
        prompt_type='list',
        **kwargs,
    )
    if user_input.quality == 'medium':
        print('  🚀 you trade-off a bit of quality for having the best speed')
    elif user_input.quality == 'good':
        print('  ⚖️ you have the best out of speed and quality')
    elif user_input.quality == 'excellent':
        print('  ✨ you trade-off speed to having the best quality')

    _, user_input.model_variant = IMAGE_MODEL_QUALITY_MAP[user_input.quality]


def _construct_local_cluster_choices(active_context, contexts):
    context_names = _get_context_names(contexts, active_context)
    choices = [NEW_CLUSTER]
    # filter contexts with `gke`
    if len(context_names) > 0 and len(context_names[0]) > 0:
        context_names = [context for context in context_names if 'gke' not in context]
        choices = context_names + choices
    return choices


def maybe_prompt_user(questions, attribute, **kwargs):
    """
    Checks the `kwargs` for the `attribute` name. If present, the value is returned directly.
    If not, the user is prompted via the cmd-line using the `questions` argument.

    :param questions: A dictionary that is passed to `PyInquirer.prompt`
        See docs: https://github.com/CITGuru/PyInquirer#documentation
    :param attribute: Name of the value to get. Make sure this matches the name in `kwargs`

    :return: A single value of either from `kwargs` or the user cli input.
    """
    if kwargs and kwargs.get(attribute) is not None:
        return kwargs[attribute]
    else:
        answer = prompt(questions)
        if attribute in answer:
            return answer[attribute]
        else:
            print("\n" * 10)
            cowsay.cow('see you soon 👋')
            exit(0)


def _prompt_value(
    name: str,
    prompt_message: str,
    prompt_type: str = 'input',
    choices: Optional[List[Union[Dict, str]]] = None,
    **kwargs: Dict,
):
    qs = {'name': name, 'type': prompt_type, 'message': prompt_message}

    if choices is not None:
        qs['choices'] = choices
        qs['type'] = 'list'
    return maybe_prompt_user(qs, name, **kwargs)


def _get_context_names(contexts, active_context=None):
    names = [c for c in contexts] if contexts is not None else []
    if active_context is not None:
        names.remove(active_context)
        names = [active_context] + names
    return names


def _cluster_running(cluster):
    config.load_kube_config(context=cluster)
    v1 = client.CoreV1Api()
    try:
        v1.list_namespace()
    except Exception as e:
        return False
    return True


def _maybe_login_wolf():
    if not os.path.exists(user('~/.jina/config.json')):
        with yaspin_extended(
            sigmap=sigmap, text='Log in to JCloud', color='green'
        ) as spinner:
            cmd('jcloud login')
        spinner.ok('🛠️')


def _parse_custom_data_from_cli(user_input: UserInput):
    data = user_input.data
    try:
        data = os.path.expanduser(data)
    except Exception:
        pass
    if os.path.exists(data):
        user_input.custom_dataset_type = DatasetTypes.PATH
        user_input.dataset_path = data
    elif 'http' in data:
        user_input.custom_dataset_type = DatasetTypes.URL
        user_input.dataset_url = data
    else:
        user_input.custom_dataset_type = DatasetTypes.DOCARRAY
        user_input.dataset_secret = data


def _handle_ffmpeg_install_required():
    bc_red = '\033[91m'
    bc_end = '\033[0m'
    print()
    print(
        f"{bc_red}To use the audio modality you need the ffmpeg audio processing"
        f" library installed on your system.{bc_end}"
    )
    print(
        f"{bc_red}For MacOS please run 'brew install ffmpeg' and on"
        f" Linux 'apt-get install ffmpeg libavcodec-extra'.{bc_end}"
    )
    print(
        f"{bc_red}After the installation, restart Jina Now and have fun with music search 🎸!{bc_end}"
    )
    cowsay.cow('see you soon 👋')
    exit(1)



import logging.config
import sys

import uvicorn
from fastapi import FastAPI

import deployment.bff.app.settings as api_settings
from deployment.bff.app.decorators import api_method, timed
from deployment.bff.app.v1.api import v1_router

logging.config.dictConfig(api_settings.DEFAULT_LOGGING_CONFIG)
logger = logging.getLogger('bff.app')
logger.setLevel(api_settings.DEFAULT_LOGGING_LEVEL)

TITLE = 'Jina NOW'
DESCRIPTION = 'The Jina NOW service API'
AUTHOR = 'Jina AI'
EMAIL = 'hello@jina.ai'
__version__ = "latest"


def build_app():
    """Build FastAPI app."""
    app = FastAPI(
        title=TITLE,
        description=DESCRIPTION,
        contact={
            'author': AUTHOR,
            'email': EMAIL,
        },
    )

    @app.get('/ping')
    @api_method
    @timed
    def check_liveness() -> str:
        """
        Sanity check - this will let the caller know that the service is operational.
        """
        return 'pong!'

    @app.get('/')
    @api_method
    @timed
    def read_root() -> str:
        """
        Root path welcome message.
        """
        return (
            f'{TITLE} v{__version__} 🚀 {DESCRIPTION} ✨ '
            f'author: {AUTHOR} email: {EMAIL} 📄  '
            'Check out /docs or /redoc for the API documentation!'
        )

    @app.on_event('startup')
    def startup():
        logger.info(
            f'Jina NOW started! ' f'Listening to [::]:{api_settings.DEFAULT_PORT}'
        )

    app.include_router(v1_router, prefix='/api/v1')

    return app


application = build_app()


def run_server():
    """Run server."""
    app = build_app()

    # start the server!
    uvicorn.run(
        app,
        host='0.0.0.0',
        port=8080,
        loop='uvloop',
        http='httptools',
    )


if __name__ == '__main__':
    try:
        run_server()
    except Exception as exc:
        logger.critical(str(exc))
        logger.exception(exc)
        sys.exit(1)



import os
import pathlib
import platform
import sys
import warnings
from os.path import expanduser as user

import cpuinfo

from now.deployment.deployment import cmd

warnings.filterwarnings("ignore")

if len(sys.argv) != 1 and not ('-h' in sys.argv[1:] or '--help' in sys.argv[1:]):
    print('Initialising Jina NOW...')
cur_dir = pathlib.Path(__file__).parents[1].resolve()

os.environ['JINA_CHECK_VERSION'] = 'False'
os.environ['JCLOUD_LOGLEVEL'] = 'ERROR'


def _get_run_args():
    from now.cli.parser import get_main_parser

    parser = get_main_parser()
    if len(sys.argv) == 1:
        parser.print_help()
        exit()
    args, unknown = parser.parse_known_args()

    if unknown:
        raise Exception('unknown args: ', unknown)

    return args


def _is_latest_version(suppress_on_error=True):
    try:
        import json
        from urllib.request import Request, urlopen

        from jina import __version__

        req = Request(
            'https://api.jina.ai/latest', headers={'User-Agent': 'Mozilla/5.0'}
        )
        with urlopen(
            req, timeout=5
        ) as resp:  # 'with' is important to close the resource after use
            latest_ver = json.load(resp)['version']
            from packaging.version import Version

            latest_ver = Version(latest_ver)
            cur_ver = Version(__version__)
            if cur_ver < latest_ver:
                from jina.logging.predefined import default_logger

                default_logger.warning(
                    f'You are using Jina version {cur_ver}, however version {latest_ver} is available. '
                    f'You should consider upgrading via the "pip install --upgrade jina" command.'
                )
                return False
        return True
    except Exception:
        if not suppress_on_error:
            raise


def _get_kind_path() -> str:
    os_type = platform.system().lower()
    # kind needs no distinction of architecture type
    kind_path, _ = cmd('which kind')
    kind_path = kind_path.strip()
    if not kind_path:
        if not os.path.exists(user('~/.cache/jina-now/kind')):
            print('kind not found. Installing kind')
            cmd(
                f'/bin/bash {cur_dir}/scripts/install_kind.sh {os_type}',
                std_output=True,
            )
        kind_path = user('~/.cache/jina-now/kind')
    else:
        kind_path = kind_path.decode('utf-8')
    return kind_path


def _get_kubectl_path() -> str:
    os_type = platform.system().lower()
    arch = 'x86_64'
    if os_type == 'darwin':
        if 'm1' in cpuinfo.get_cpu_info().get('brand_raw').lower():
            arch = 'arm64'
        else:
            arch = platform.machine()
    elif os_type == 'linux':
        arch = platform.machine()

    # kubectl needs `intel` or `m1` for apple os
    # for linux no need of architecture type
    kubectl_path, _ = cmd('which kubectl')
    kubectl_path = kubectl_path.strip()
    if not kubectl_path:
        if not os.path.isfile(user('~/.cache/jina-now/kubectl')):
            print('kubectl not found. Installing kubectl')
            cmd(
                f'/bin/bash {cur_dir}/scripts/install_kubectl.sh {os_type} {arch}',
                std_output=True,
            )
        kubectl_path = user('~/.cache/jina-now/kubectl')
    else:
        kubectl_path = kubectl_path.decode('utf-8')
    return kubectl_path


def cli(args=None):
    """The main entrypoint of the CLI"""
    if not args:
        args = _get_run_args()

    if '--version' in sys.argv[1:]:
        from now import __version__

        print(__version__)
        exit(0)

    os.environ['JINA_LOG_LEVEL'] = 'CRITICAL'
    os_type = platform.system().lower()
    arch = 'x86_64'
    if os_type == 'darwin':
        if 'm1' in cpuinfo.get_cpu_info().get('brand_raw').lower():
            arch = 'arm64'
        else:
            arch = platform.machine()
    elif os_type == 'linux':
        arch = platform.machine()
    args = vars(args)  # Make it a dict from Namespace

    args['kubectl_path'] = _get_kubectl_path()
    args['kind_path'] = _get_kind_path()
    from now.run_all_k8s import run_k8s

    run_k8s(os_type=os_type, arch=arch, **args)


if __name__ == '__main__':
    cli()



from fastapi import APIRouter

from deployment.bff.app.v1.routers import image, text

v1_router = APIRouter()
v1_router.include_router(image.router, prefix='/image', tags=['Image'])
v1_router.include_router(text.router, prefix='/text', tags=['Text'])



import base64
from typing import List

from docarray import Document, DocumentArray
from fastapi import APIRouter
from jina import Client

from deployment.bff.app.v1.models.image import (
    NowImageIndexRequestModel,
    NowImageResponseModel,
    NowImageSearchRequestModel,
)
from deployment.bff.app.v1.routers.helper import process_query

router = APIRouter()


# Index
@router.post(
    "/index",
    summary='Add more data to the indexer',
)
def index(data: NowImageIndexRequestModel):
    """
    Append the list of image data to the indexer. Each image data should be
    `base64` encoded using human-readable characters - `utf-8`.
    """
    index_docs = DocumentArray()
    for image in data.images:
        base64_bytes = image.encode('utf-8')
        message = base64.decodebytes(base64_bytes)
        index_docs.append(Document(blob=message))

    if 'wolf.jina.ai' in data.host:
        c = Client(host=data.host)
    else:
        c = Client(host=data.host, port=data.port)
    c.post('/index', index_docs)


# Search
@router.post(
    "/search",
    response_model=List[NowImageResponseModel],
    summary='Search image data via text or image as query',
)
def search(data: NowImageSearchRequestModel):
    """
    Retrieve matching images for a given query. Image query should be `base64` encoded
    using human-readable characters - `utf-8`.
    """
    query_doc = process_query(data.text, data.image)
    if 'wolf.jina.ai' in data.host:
        c = Client(host=data.host)
    else:
        c = Client(host=data.host, port=data.port)
    docs = c.post('/search', query_doc, parameters={"limit": data.limit})
    return docs[0].matches.to_dict()



import base64
import os
import uuid
from copy import deepcopy
from os.path import join as osp
from pathlib import Path
from typing import Optional

from docarray import Document, DocumentArray

from now.constants import (
    BASE_STORAGE_URL,
    IMAGE_MODEL_QUALITY_MAP,
    DatasetTypes,
    Modalities,
    Qualities,
)
from now.data_loading.convert_datasets_to_jpeg import to_thumbnail_jpg
from now.dialog import UserInput
from now.log.log import yaspin_extended
from now.utils import download, sigmap


def load_data(user_input: UserInput) -> DocumentArray:
    """
    Based on the user input, this function will pull the configured DocArray.

    :param user_input: The configured user object. Result from the Jina Now cli dialog.
    :return: The loaded DocumentArray.
    """
    da = None

    if not user_input.is_custom_dataset:
        print('⬇  Download DocArray')
        url = get_dataset_url(
            user_input.data, user_input.quality, user_input.output_modality
        )
        da = _fetch_da_from_url(url)

    else:
        if user_input.custom_dataset_type == DatasetTypes.DOCARRAY:
            print('⬇  Pull DocArray')
            da = _pull_docarray(user_input.dataset_secret)
        elif user_input.custom_dataset_type == DatasetTypes.URL:
            print('⬇  Pull DocArray')
            da = _fetch_da_from_url(user_input.dataset_url)
        elif user_input.custom_dataset_type == DatasetTypes.PATH:
            print('💿  Loading DocArray from disk')
            da = _load_from_disk(user_input.dataset_path, user_input.output_modality)

    if da is None:
        raise ValueError(
            f'Could not load DocArray. Please check your configuration: {user_input}.'
        )
    da = da.shuffle(seed=42)
    da = _deep_copy_da(da)
    return da


def _fetch_da_from_url(
    url: str, downloaded_path: str = '~/.cache/jina-now'
) -> DocumentArray:
    data_dir = os.path.expanduser(downloaded_path)
    if not os.path.exists(osp(data_dir, 'data/tmp')):
        os.makedirs(osp(data_dir, 'data/tmp'))
    data_path = (
        data_dir
        + f"/data/tmp/{base64.b64encode(bytes(url, 'utf-8')).decode('utf-8')}.bin"
    )
    if not os.path.exists(data_path):
        download(url, data_path)

    with yaspin_extended(
        sigmap=sigmap, text="Extracting dataset", color="green"
    ) as spinner:
        da = DocumentArray.load_binary(data_path)
        spinner.ok("📂")
    return da


def _pull_docarray(dataset_secret: str):
    try:
        return DocumentArray.pull(token=dataset_secret, show_progress=True)
    except Exception:
        print(
            '💔 oh no, the secret of your docarray is wrong, or it was deleted after 14 days'
        )
        exit(1)


def _load_from_disk(dataset_path: str, modality: Modalities) -> DocumentArray:
    if os.path.isfile(dataset_path):
        try:
            return DocumentArray.load_binary(dataset_path)
        except Exception as e:
            print(f'Failed to load the binary file provided under path {dataset_path}')
            exit(1)
    elif os.path.isdir(dataset_path):
        with yaspin_extended(
            sigmap=sigmap, text="Loading and pre-processing data", color="green"
        ) as spinner:
            if modality == Modalities.IMAGE:
                return _load_images_from_folder(dataset_path)
            elif modality == Modalities.TEXT:
                return _load_texts_from_folder(dataset_path)
            elif modality == Modalities.MUSIC:
                return _load_music_from_folder(dataset_path)
            spinner.ok('🏭')
    else:
        raise ValueError(
            f'The provided dataset path {dataset_path} does not'
            f' appear to be a valid file or folder on your system.'
        )


def _load_images_from_folder(path: str) -> DocumentArray:
    def convert_fn(d):
        try:
            d.load_uri_to_image_tensor()
            return to_thumbnail_jpg(d)
        except:
            return d

    da = DocumentArray.from_files(path + '/**')
    da.apply(convert_fn)
    return DocumentArray(d for d in da if d.blob != b'')


def _load_music_from_folder(path: str):
    from pydub import AudioSegment

    def convert_fn(d: Document):
        try:
            AudioSegment.from_file(d.uri)  # checks if file is valid
            with open(d.uri, 'rb') as fh:
                d.blob = fh.read()
            return d
        except Exception as e:
            return d

    da = DocumentArray.from_files(path + '/**')
    da.apply(convert_fn)
    return DocumentArray(d for d in da if d.blob != b'')


def _load_texts_from_folder(path: str) -> DocumentArray:
    import nltk

    nltk.download('punkt', quiet=True)
    from nltk.tokenize import sent_tokenize

    def convert_fn(d):
        try:
            d.load_uri_to_text()
            d.tags['additional_info'] = str(Path(d.uri).relative_to(path))
            return d
        except:
            return d

    def split_document(d):
        return DocumentArray(
            (
                Document(
                    mime_type='text',
                    text=sentence,
                    tags=d.tags,
                )
                for sentence in set(sent_tokenize(d.text.replace('\n', ' ')))
            )
        )

    da = DocumentArray.from_files(path + '/*.txt')
    da.apply(convert_fn)

    ret = DocumentArray()
    for d in da:
        ret += split_document(d)
    return ret


def get_dataset_url(
    dataset: str, model_quality: Optional[Qualities], output_modality: Modalities
) -> str:
    data_folder = None
    if output_modality == Modalities.IMAGE:
        data_folder = 'jpeg'
    elif output_modality == Modalities.TEXT:
        data_folder = 'text'
    elif output_modality == Modalities.MUSIC:
        data_folder = 'music'

    if model_quality is not None:
        return f'{BASE_STORAGE_URL}/{data_folder}/{dataset}.{IMAGE_MODEL_QUALITY_MAP[model_quality][0]}.bin'
    else:
        return f'{BASE_STORAGE_URL}/{data_folder}/{dataset}.bin'


def _deep_copy_da(da: DocumentArray) -> DocumentArray:
    new_da = DocumentArray()
    for i, d in enumerate(da):
        new_doc = deepcopy(d)
        new_doc.id = str(uuid.uuid4())
        new_da.append(new_doc)
    return new_da



import argparse

from jina.parsers.helper import _ColoredHelpFormatter

from now import __version__


def set_base_parser():
    """Set the base parser
    :return: the parser
    """

    # create the top-level parser
    urls = {
        'Code': ('💻', 'https://github.com/jina-ai/now'),
        'Jina Docs': ('📖', 'https://docs.jina.ai'),
        'Help': ('💬', 'https://slack.jina.ai'),
        'Hiring!': ('🙌', 'https://career.jina.ai'),
    }
    url_str = '\n'.join(f'- {v[0]:<10} {k:10.10}\t{v[1]}' for k, v in urls.items())

    parser = argparse.ArgumentParser(
        epilog=f'Jina NOW - get your neural search case up in minutes. \n\n {url_str}',
        formatter_class=_chf,
        description='Command Line Interface of `%(prog)s`',
    )
    parser.add_argument(
        '-v',
        '--version',
        action='version',
        version=__version__,
        help='Show Jina version',
    )
    return parser


def set_help_parser(parser=None):
    """Set the parser for the jina help lookup
    :param parser: an optional existing parser to build upon
    :return: the parser
    """

    if not parser:
        from jina.parsers.base import set_base_parser

        parser = set_base_parser()

    parser.add_argument(
        'query',
        type=str,
        help='Lookup the usage & mention of the argument name in Jina NOW',
    )
    return parser


def set_start_parser(sp=None):
    """Add the arguments for the jina now to the parser
    :param parser: an optional existing parser to build upon
    :return: the parser
    """

    parser = sp.add_parser(
        'start',
        help='Start jina now and create or reuse a cluster.',
        description='Start jina now and create or reuse a cluster.',
        formatter_class=_chf,
    )

    parser.add_argument(
        '--output_modality',
        help='Select the modality you want to develop a search flow for.',
        type=str,
    )

    parser.add_argument(
        '--data',
        help='Select one of the available datasets or provide local filepath, '
        'docarray url, or docarray secret to use your own dataset',
        type=str,
    )

    parser.add_argument(
        '--quality',
        help='Choose the quality of the model that you would like to finetune',
        type=str,
    )

    parser.add_argument(
        '--cluster',
        help='Reference an existing `local` cluster or select `new` to create a new one.',
        type=str,
    )

    parser.add_argument(
        '--deployment-type',
        help='Option is `local` and `remote`. Select `local` if you want search engine to be deployed on local '
        'cluster. Select `remote` to deploy it on Jina Cloud',
        type=str,
    )


def set_stop_parser(sp):
    sp.add_parser(
        'stop',
        help='Stop jina now and remove local cluster.',
        description='Stop jina now and remove local cluster.',
        formatter_class=_chf,
    )


def set_survey_parser(sp):
    sp.add_parser(
        'survey',
        help='Opens a survey in the browser. Thanks for providing us feedback.',
        description='Opens a survey in the browser. Thanks for providing us feedback.',
        formatter_class=_chf,
    )


def get_main_parser():
    """The main parser for Jina NOW
    :return: the parser
    """
    # create the top-level parser
    parser = set_base_parser()
    sp = parser.add_subparsers(
        dest='cli',
        description='',
        required=True,
    )
    set_start_parser(sp)
    set_stop_parser(sp)
    set_survey_parser(sp)

    return parser


_chf = _ColoredHelpFormatter
